/*****************************************************************************
   gtkAbsinthe - Front End for the Absinthe Core Library
   This software is Copyright (C) 2004  nummish, 0x90.org

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/
#define DEBUG

using System;
using System.Configuration;
using System.Threading;

using Gtk;
using Glade;
using GLib;
using Gdk;
using GtkSharp;

using System.Collections;
using System.Collections.Specialized;
using System.Text.RegularExpressions;
using System.Net;

using Absinthe.Core;

public class MainWin
{

	private Hashtable InjectionFields = new Hashtable(); 
	private DataStore _SaveInfo;
	private Absinthe.LocalSettings _AppSettings = new Absinthe.LocalSettings(); // This is App specific settings

	delegate void ThreadedSub();
	ThreadNotify _GuiAction = null;
	Queue _ActionQueue;
	string[] _PluginEntries;

	// {{{ Glade Widgets
	// Bind the M^HWidgets
	[Glade.Widget] Gtk.Window Absinthe;
	private Glade.XML MainGladeXml;
	[Glade.Widget] Gtk.Button butScan;
	[Glade.Widget] Gtk.Statusbar myStatusBar;
	[Glade.Widget] Gtk.Dialog dlgError;
	[Glade.Widget] Gtk.Window AboutForm;
	[Glade.Widget] Gtk.Image AboutImage;
	[Glade.Widget] Gtk.Button butAboutOk;
	[Glade.Widget] Gtk.Label lblErrorMsg;
	[Glade.Widget] Gtk.Button okbutton1;

	// Tab 1
	[Glade.Widget] Gtk.Entry txtTargetURL;
	[Glade.Widget] Gtk.RadioButton optBlindInjection;
	[Glade.Widget] Gtk.RadioButton optErrorBasedInjection;
	[Glade.Widget] Gtk.Combo cboPluginList;
	
	[Glade.Widget] Gtk.RadioButton optConnectGet;
	[Glade.Widget] Gtk.RadioButton optConnectPost;
	[Glade.Widget] Gtk.CheckButton chkUseSSL;
	[Glade.Widget] Gtk.CheckButton chkTerminateQuery;
	[Glade.Widget] Gtk.Button butConfigProxy;
	[Glade.Widget] Gtk.CheckButton chkAppendExtraQuery;
	[Glade.Widget] Gtk.Label lblAppendExtraQuery;
	[Glade.Widget] Gtk.Entry txtAppendExtraQuery;

	[Glade.Widget] Gtk.Entry txtParamName;
	[Glade.Widget] Gtk.Entry txtParamValue;
	[Glade.Widget] Gtk.CheckButton chkInjectable;
	[Glade.Widget] Gtk.CheckButton chkTreatAsString;
	[Glade.Widget] Gtk.Button butAddParam;
	[Glade.Widget] Gtk.Button butAddHeader;
	[Glade.Widget] Gtk.Notebook tabParams;
	[Glade.Widget] Gtk.TreeView lstParams;
	[Glade.Widget] Gtk.Button butEditParam;
	[Glade.Widget] Gtk.Button butDeleteParam;
	[Glade.Widget] Gtk.TreeView lstCookies;

	
	// Schema Tab
	[Glade.Widget] Gtk.Button butGetUsername;
	[Glade.Widget] Gtk.Button butLoadFieldInfo;
	[Glade.Widget] Gtk.Button butLoadTableInfo;
	[Glade.Widget] Gtk.TreeView lstSchema;
	
	// Proxy Config Widgets
	[Glade.Widget] Gtk.Window ProxyConfig;
	[Glade.Widget] Gtk.CheckButton chkUseProxy;
	[Glade.Widget] Gtk.Entry txtProxyAddress;
	[Glade.Widget] Gtk.Entry txtProxyPort;
	[Glade.Widget] Gtk.Button butSaveProxySettings;
	[Glade.Widget] Gtk.Button butCancelProxySettings;
	[Glade.Widget] Gtk.TreeView lstProxyList;
	[Glade.Widget] Gtk.Button butAddProxy;
	[Glade.Widget] Gtk.Button butRemoveProxy;

	// File Dialog
	[Glade.Widget] Gtk.FileSelection FileSelectForm;
	[Glade.Widget] Gtk.Button butFileSelectCancel;
	[Glade.Widget] Gtk.Button butFileSelectOk;

	// Pull Data Tab
	[Glade.Widget] Gtk.TreeView lstAvailFields;
	[Glade.Widget] Gtk.TreeView lstSelectedFields;
	[Glade.Widget] Gtk.Entry txtPullDataXml;

	// }}}

	// {{{ GuiActions Enum
	private enum GuiActions : byte
	{
		StatusMessage = 1,
		ProgressIncrease = 2,
		ProgressClear = 3,
		ControlEnableSwitch = 4,
		ControlVisibleSwitch = 5
	}
	// }}}

	// {{{ FileAction Enum
	private enum FileAction : byte
	{
		SaveAs = 1,
		Open = 2,
		Pulldata = 3
	}
	// }}}

	// {{{ MainWin
	public MainWin() 
	{
		Application.Init();

		MainGladeXml = new Glade.XML (null, "gtkabsinthe.glade", "Absinthe", null);
		//MainGladeXml = new Glade.XML ("gtkabsinthe.glade", "Absinthe", null);
		MainGladeXml.Autoconnect(this);
		this.Absinthe = (Gtk.Window) MainGladeXml.GetWidget("Absinthe");
		_SaveInfo = new DataStore(new GlobalDS.OutputStatusDelegate(SafeOutput));
		
		//ReadSettingsFile();

		InitUI();

		Gtk.Application.Run ();
	}
	// }}}

	// {{{ GenerateErrorDialog
	private void GenerateErrorDialog(string ErrorMessage)
	{
		
		Glade.XML gxml = new Glade.XML (null, "gtkabsinthe.glade", "dlgError", null);
		//Glade.XML gxml = new Glade.XML ("gtkabsinthe.glade", "dlgError", null);
		dlgError = (Gtk.Dialog) gxml.GetWidget("dlgError");
		lblErrorMsg = (Gtk.Label) gxml.GetWidget("lblErrorMsg");
		lblErrorMsg.Text = ErrorMessage;
		okbutton1 = (Gtk.Button) gxml.GetWidget("okbutton1");
		okbutton1.Clicked += new EventHandler (ErrorOk_Click);
	}
	// }}}

	// {{{ InitUI
	private void InitUI()
	{
		chkUseSSL.Sensitive = false;

		// Demo TreeList shit
		ListStore store = new ListStore (typeof(string), typeof(string), typeof(bool), typeof(bool));

		lstParams.Model = store;
		lstParams.HeadersVisible = true;

		if (lstParams.Columns.Length == 0)
		{
			lstParams.AppendColumn ("Name", new CellRendererText(), "text", 0);
			lstParams.AppendColumn ("Default", new CellRendererText(), "text", 1);
			lstParams.AppendColumn ("Injectable", new CellRendererToggle(), "active", 2);
			lstParams.AppendColumn ("String", new CellRendererToggle(), "active", 3);
		}

		ListStore cstore = new ListStore(typeof(string), typeof(string));
		lstCookies.Model = cstore;
		lstCookies.HeadersVisible = true;

		if (lstCookies.Columns.Length == 0)
		{
			lstCookies.AppendColumn ("Name", new CellRendererText(), "text", 0);
			lstCookies.AppendColumn ("Value", new CellRendererText(), "text", 1);
		}
		
		_GuiAction = new ThreadNotify(new ReadyEvent(TakeGuiAction));
		_ActionQueue = new Queue();

		InitSchemaTreeView();

		if (lstAvailFields.Columns.Length == 0)
		{
			lstAvailFields.AppendColumn("Table", new CellRendererText(), "text", 0);
			lstAvailFields.AppendColumn("Field", new CellRendererText(), "text", 1);
		}

		ListStore fstore = new ListStore (typeof(string), typeof(string), typeof(int));
		lstSelectedFields.Model = fstore;
		
		if (lstSelectedFields.Columns.Length == 0)
		{
			lstSelectedFields.AppendColumn("Table", new CellRendererText(), "text", 0);
			lstSelectedFields.AppendColumn("Field", new CellRendererText(), "text", 1);
		}

		LoadPluginList();

		chkTreatAsString.Sensitive = false;
	}
	// }}}

	// {{{ LoadPluginList
	private void LoadPluginList()
	{
		ArrayList al = _SaveInfo.PluginList;
		ArrayList NameList = new ArrayList();

		foreach (PluginTemplate pt in al)
		{
			if (!NameList.Contains(pt.PluginDisplayTargetName))
			{
				NameList.Add(pt.PluginDisplayTargetName);
			}
			else
			{
				string EntryName = pt.PluginDisplayTargetName;
				string Modifier = ""; int ModCounter = 0;
				while (NameList.Contains(EntryName + Modifier))
				{
					ModCounter++;
					Modifier = " {" + ModCounter + "}";
				}
				NameList.Add(EntryName + Modifier);
			}
		}

		_PluginEntries = (string[]) NameList.ToArray(typeof(string));
		cboPluginList.PopdownStrings = _PluginEntries;
	}
	// }}}

	// {{{ InitSchemaTreeView
	private void InitSchemaTreeView()
	{
		TreeStore store = new TreeStore(typeof(string), typeof(string));
		
		TreeIter iter, db_iter;
		
		db_iter = store.AppendValues("Database", "");
		iter = store.AppendValues(db_iter, "Username", "??? UNKNOWN ???");
		iter = store.AppendValues(db_iter, "Tables", "??? UNKNOWN ???");
		lstSchema.Model = store;
		lstSchema.HeadersVisible = true;

		if (lstSchema.Columns.Length == 0)
		{
			lstSchema.AppendColumn ("Name", new CellRendererText(), "text", 0);
			lstSchema.AppendColumn ("Value", new CellRendererText(), "text", 1);
		}

	}
	// }}}

	// {{{ LoadFieldDataFromTableName
	private void LoadFieldDataFromTableName(string TableName, TreeIter TableIter, bool Disconnected)
	{
		TreeIter FieldIter;
		TreeStore store = (TreeStore) lstSchema.Model;

		store.IterNthChild(out FieldIter, TableIter, 2);

		for (int i=0; i < _SaveInfo.TableList.Length; i++)
		{
			if (_SaveInfo.TableList[i].Name.Equals(TableName))
			{
				// If it's already been loaded, clear it before you reload it
				store.Remove(ref FieldIter);
				FieldIter = store.AppendValues(TableIter, "Fields", "??? UNKNOWN ???");

				PopulateFieldView(i, FieldIter, Disconnected);
			}
		}
	}
	// }}}

	// {{{ PopulateFieldView
	private void PopulateFieldView(int TableIndex, TreeIter FieldIter, bool Disconnected)
	{
		if (_SaveInfo.TableList[TableIndex].FieldCount == 0)
		{
			if (Disconnected) return;
			
			_SaveInfo.TargetAttackVector.PopulateTableStructure(ref (_SaveInfo.TableList[TableIndex]));
		}

		GlobalDS.Field[] FieldList = _SaveInfo.TableList[TableIndex].FieldList;
		string TableName = _SaveInfo.TableList[TableIndex].Name;
		TreeIter iter;

		TreeStore store = (TreeStore) lstSchema.Model;
		store.SetValue(FieldIter, 1, _SaveInfo.TableList[TableIndex].FieldCount.ToString());

		
		for (int i = 0; i < _SaveInfo.TableList[TableIndex].FieldCount; i++)
		{
			string FieldName = FieldList[i].FieldName;
			iter = store.AppendValues(FieldIter, FieldName, FieldList[i].DataType.ToString());
			// Can we underline primary keys?
		}

		ReloadAvailableFields();
	}
	// }}}

	// {{{ ReloadAvailableFields
	private void ReloadAvailableFields()
	{
		ListStore fstore = new ListStore (typeof(string), typeof(string), typeof(int));
		ListStore selstore = (ListStore) lstSelectedFields.Model;

		TreeIter SelIter;
		Hashtable selected = new Hashtable();

		if(selstore.GetIterFirst(out SelIter))
		{
			string FullName = (string) selstore.GetValue(SelIter, 0) + "." + (string) selstore.GetValue(SelIter, 1);

			selected.Add(FullName, 0);
			
			while (selstore.IterNext(ref SelIter))
			{
				FullName = (string) selstore.GetValue(SelIter, 0) + "." + (string) selstore.GetValue(SelIter, 1);

				selected.Add(FullName, 0);
			}
		}
		
		for (int i = 0; i < _SaveInfo.TableList.Length; i++)
		{
			if (_SaveInfo.TableList[i].FieldCount > 0)
			{
				GlobalDS.Field[] FieldList = _SaveInfo.TableList[i].FieldList;
				for (int j = 0; j < _SaveInfo.TableList[i].FieldCount; j++)
				{
					string FieldName = FieldList[j].FieldName;
					if (!selected.ContainsKey(_SaveInfo.TableList[i].Name +"."+ FieldName))
					{
						fstore.AppendValues(_SaveInfo.TableList[i].Name,FieldName, j);
					}
				}
			}
		}

		lstAvailFields.Model = fstore;
	}
	// }}}

	// {{{ RetrieveTableInfoFromDatabase
	private void RetrieveTableInfoFromDatabase()
	{
		try
		{
			SafeChangeSensitivity(butLoadTableInfo, false);
			SafeStatusBarWrite("Gathering Table Information");
			if (_SaveInfo.TargetAttackVector != null)
			{
				_SaveInfo.TableList = _SaveInfo.TargetAttackVector.GetTableList();
				ThreadNotify tn = new ThreadNotify(new ReadyEvent(AdjustTableInfoGuiInThread));
				tn.WakeupMain();
			}
			SafeStatusBarWrite("Finished Gathering Table Information");
		}
		catch (Exception e)
		{
			System.Console.WriteLine(e.ToString());
		//	GenerateErrorDialog("Table information could not be retrieved");
		}
		finally
		{
			SafeChangeSensitivity(butLoadTableInfo, true);
		}
	}
	// }}}

	// {{{ AdjustTableInfoGuiInThread
	private void AdjustTableInfoGuiInThread()
	{
		butLoadFieldInfo.Visible = true;
		LoadTableInfoList();
	}
	// }}}

	// {{{ LoadTableInfoList
	private void LoadTableInfoList()
	{
		TreeIter iter, db_iter, TableIter;

		if (_SaveInfo.TableList != null)
		{
			// Load the treeview
			TreeStore tstore = (TreeStore) lstSchema.Model;
			tstore.GetIterFirst(out db_iter);
			tstore.IterChildren(out iter, db_iter);
			tstore.IterNext(ref iter);
			tstore.SetValue(iter, 1, _SaveInfo.TableList.Length.ToString() + " Tables");


			for (int i=0; i < _SaveInfo.TableList.Length; i++)
			{
				/*store.AppendValues(_SaveInfo.TableList[i].Name, _SaveInfo.TableList[i].RecordCount.ToString());*/

				TableIter = tstore.AppendValues(iter,  _SaveInfo.TableList[i].Name, string.Empty);
				tstore.AppendValues(TableIter, "ID", _SaveInfo.TableList[i].ObjectID.ToString());
				tstore.AppendValues(TableIter, "Record Count", _SaveInfo.TableList[i].RecordCount.ToString());
				tstore.AppendValues(TableIter, "Fields", "??? UNKNOWN ???");
			}

			/*lstTableInfo.Model = store;*/
		}
	}
	// }}}

	// {{{ RetrieveUsernameFromDatabase
	private void RetrieveUsernameFromDatabase()
	{
		try
		{
			SafeChangeSensitivity(butGetUsername, false);
			SafeStatusBarWrite("Retrieving username");

			if (_SaveInfo.TargetAttackVector != null)
			{
				SafeOutput("mrh");
				if (_SaveInfo.TargetAttackVector == null) SafeOutput("nooooo");
				_SaveInfo.Username = _SaveInfo.TargetAttackVector.GetDatabaseUsername();
				SafeOutput("hrm");
				ThreadNotify tn = new ThreadNotify(new ReadyEvent(AdjustDbLoginText));
				tn.WakeupMain();
			}
			else
			{
				SafeOutput("The attack vector is null jerkface");
			}
			/* SafeChangeVisibility(txtDBlogin, true);*/
		}
		catch (Exception e)
		{
			throw e;
		}
		finally
		{
			SafeChangeSensitivity(butGetUsername, true);
			SafeStatusBarWrite("Username retrieved");
		}
	}
	// }}}

	// {{{ AdjustDbLoginText
	private void AdjustDbLoginText()
	{
		TreeIter db_iter, iter;
		
		/*txtDBlogin.Text = _SaveInfo.Username;*/
		TreeStore store = (TreeStore) lstSchema.Model;
		store.GetIterFirst(out db_iter);
		store.IterChildren(out iter, db_iter);
		store.SetValue(iter, 1, _SaveInfo.Username);
	}
	// }}}

	// {{{ OnAbsintheDelete
	public void OnAbsintheDelete(object sender, EventArgs args)
	{
		Application.Quit();
	}
	// }}}

	// {{{ ErrorOk_Click
	public void ErrorOk_Click(object sender, EventArgs args)
	{
		dlgError.Destroy();
	}
	// }}}

	// {{{ ConnectMethod
	private string ConnectMethod()
	{
		string Method;

		if (optConnectPost.Active == true)
		{
			Method = "Post";
		}
		else if (optConnectGet.Active == true)
		{
			Method = "Get";
		}
		else
		{
			//myStatusBar.Push(myStatusBar.GetContextId("Connection_Error"), "  No connection method selected");
			return "";
		}

		return Method;
	}
	// }}}

	// {{{ TakeGuiAction
	private void TakeGuiAction()
	{
		Gtk.Widget tmp;
			lock (_ActionQueue)
		{
				while (_ActionQueue.Count > 0)
				{
					switch ((GuiActions) _ActionQueue.Dequeue())
					{
						case GuiActions.StatusMessage:
							string Msg = _ActionQueue.Dequeue().ToString();
							myStatusBar.Push(myStatusBar.GetContextId("regular"), Msg);
							break;
						case GuiActions.ProgressIncrease:
							break;
						case GuiActions.ProgressClear:
							break;
						case GuiActions.ControlEnableSwitch:
							tmp = (Gtk.Widget) _ActionQueue.Dequeue();
							tmp.Sensitive = (bool) _ActionQueue.Dequeue();
							break;
						case GuiActions.ControlVisibleSwitch:
							tmp = (Gtk.Widget) _ActionQueue.Dequeue();
							tmp.Visible = (bool) _ActionQueue.Dequeue();
							break;
					}
				}
		}
	}
	// }}}

	// {{{ SafeStatusBarWrite
	private void SafeStatusBarWrite(string Message)
	{
		lock (_ActionQueue)
		{
			_ActionQueue.Enqueue(GuiActions.StatusMessage);
			_ActionQueue.Enqueue(Message);
		}

		lock (this)
		{
			_GuiAction.WakeupMain();
		}
	}
	// }}}

	// {{{ SafeChangeSensitivity
	private void SafeChangeSensitivity(Gtk.Widget GuiControl, bool Sensitive)
	{
		lock (_ActionQueue)
		{ 
			_ActionQueue.Enqueue(GuiActions.ControlEnableSwitch);
			_ActionQueue.Enqueue(GuiControl); 
			_ActionQueue.Enqueue(Sensitive);
		}

		lock (this)
		{
			_GuiAction.WakeupMain();
		}
	}
	// }}}

	// {{{ SafeChangeVisibility
	private void SafeChangeVisibility(Gtk.Widget GuiControl, bool Visible)
	{
		lock (_ActionQueue)
		{ 
			_ActionQueue.Enqueue(GuiActions.ControlVisibleSwitch);
			_ActionQueue.Enqueue(GuiControl); 
			_ActionQueue.Enqueue(Visible);
		}

		lock (this)
		{
			_GuiAction.WakeupMain();
		}
	}
	// }}}

	private void SafeOutput(string Message)
	{
		// Removed from release version
#if DEBUGISON
		System.Console.WriteLine("SAFER: {0}", Message);
#endif
	}

	// {{{ InitializeAttackVectors
	private void InitializeAttackVectors()
	{
		string URL;

		if (chkUseSSL.Active) URL = "https://";
		else URL = "http://";
		URL += txtTargetURL.Text;

		string Method = ConnectMethod();

		if (Method.Equals("")) return;

		// Generate StringDict
		string TargetName, TargetField;
		TargetName = String.Empty; TargetField = String.Empty;

		StringDictionary Others = new StringDictionary();
		StringDictionary Cookies = new StringDictionary();

		Others = ExtractFormParameters(ref TargetName, ref TargetField);
		Cookies = ExtractCookies();

		if (TargetName.Equals(String.Empty))
		{
			SafeStatusBarWrite("No Injection Point Found");
			return;
		}

		SafeStatusBarWrite("Beginning Preliminary Scan");

		try
		{
			SafeChangeSensitivity(butScan, false);

			AttackVectorFactory avf;
			GlobalDS.OutputStatusDelegate safeout = new GlobalDS.OutputStatusDelegate(SafeOutput);

			SafeOutput(chkTerminateQuery.Active.ToString());

			GlobalDS.InjectionOptions opts = new GlobalDS.InjectionOptions();
			opts.Cookies = Cookies;
			opts.TerminateQuery = chkTerminateQuery.Active;
			opts.Tolerance = _SaveInfo.FilterTolerance;
			opts.WebProxies = _AppSettings.ProxyQueue;
			opts.InjectAsString = _SaveInfo.InjectAsString;
			opts.Delimiter = _SaveInfo.FilterDelimiter;
			System.Console.WriteLine("Filter Delim: {0} && {1}", _SaveInfo.FilterDelimiter, opts.Delimiter);

			avf = new AttackVectorFactory(URL, TargetName, TargetField, Others, Method, safeout, opts);

			PluginTemplate pt = (PluginTemplate) _SaveInfo.PluginList[Array.IndexOf(_PluginEntries, cboPluginList.Entry.Text)];

			if (optBlindInjection.Active)
			{
				_SaveInfo.TargetAttackVector = avf.BuildBlindSqlAttackVector(_SaveInfo.FilterTolerance, pt);
			}
#if FULL_RELEASE
			else if(optErrorBasedInjection.Active)
			{
				_SaveInfo.TargetAttackVector = avf.BuildSqlErrorAttackVector();
			}
#endif
			SafeStatusBarWrite("Finished initial scan");
		}
		catch (Exception e)
		{
			Console.WriteLine(e.ToString());
			SafeStatusBarWrite(e.Message);
		}
		finally
		{
			SafeChangeSensitivity(butScan, true);
		}
	}
	// }}}

	// {{{ ExtractFormParameters
	private StringDictionary ExtractFormParameters(ref string TargetName, ref string TargetField)
	{
		bool FoundTarget = false;
		ListStore store = (ListStore) lstParams.Model;	
		TreeIter iter;
		string Name, Val;
		bool Injectable, AsString;
		StringDictionary retVal = new StringDictionary();

		if(store.GetIterFirst(out iter))
		{

			Name = (string) store.GetValue(iter, 0);
			Val = (string) store.GetValue(iter, 1);
			Injectable = (bool) store.GetValue(iter, 2);
			AsString = (bool) store.GetValue(iter, 3);

			if (Injectable)
			{
				FoundTarget = true;
				TargetName = Name;
				TargetField = Val;
				_SaveInfo.InjectAsString = AsString;
			}
			else
			{
				retVal[Name] = Val;
			}


			while (store.IterNext(ref iter))
			{
				Name = (string) store.GetValue(iter, 0);
				Val = (string) store.GetValue(iter, 1);
				Injectable = (bool) store.GetValue(iter, 2);
				AsString = (bool) store.GetValue(iter,3);

				if (!FoundTarget && Injectable)
				{
					FoundTarget = true;
					TargetName = Name;
					TargetField = Val;
					_SaveInfo.InjectAsString = AsString;
				}
				else
				{
					retVal[Name] = Val;
				}
			}
		}

		return retVal;
	}
	// }}}

	// {{{ ExtractCookies
	private StringDictionary ExtractCookies()
	{
		ListStore store = (ListStore) lstCookies.Model;	
		TreeIter iter;
		string Name, Val;
		StringDictionary retVal = new StringDictionary();

		if(store.GetIterFirst(out iter))
		{

			Name = (string) store.GetValue(iter, 0);
			Val = (string) store.GetValue(iter, 1);
			retVal[Name] = Val;

			while (store.IterNext(ref iter))
			{
				Name = (string) store.GetValue(iter, 0);
				Val = (string) store.GetValue(iter, 1);
				retVal[Name] = Val;
			}
		}

		return retVal;
	}
	// }}}

	// {{{ OnUseProxy_Toggle
	public void OnUseProxy_Toggle(object sender, EventArgs args)
	{
		txtProxyAddress.Sensitive = chkUseProxy.Active;
		txtProxyPort.Sensitive = chkUseProxy.Active;
		lstProxyList.Sensitive = chkUseProxy.Active;
		butAddProxy.Sensitive = chkUseProxy.Active;
		butRemoveProxy.Sensitive = chkUseProxy.Active;
	}
	// }}}

	// {{{ ConnectProxyConfigObjects
	private void ConnectProxyConfigObjects(ref Glade.XML gxml)
	{
		ProxyConfig = (Gtk.Window) gxml.GetWidget("ProxyConfig");
		chkUseProxy = (Gtk.CheckButton) gxml.GetWidget("chkUseProxy");
		txtProxyAddress = (Gtk.Entry) gxml.GetWidget("txtProxyAddress");
		txtProxyPort = (Gtk.Entry) gxml.GetWidget("txtProxyPort");
		butSaveProxySettings = (Gtk.Button) gxml.GetWidget("butSaveProxySettings");
		butCancelProxySettings = (Gtk.Button) gxml.GetWidget("butCancelProxySettings");
		butAddProxy = (Gtk.Button) gxml.GetWidget("butAddProxy");
		butRemoveProxy = (Gtk.Button) gxml.GetWidget("butRemoveProxy");
		lstProxyList = (Gtk.TreeView) gxml.GetWidget("lstProxyList");

		butAddProxy.Clicked += new EventHandler (OnAddProxy_Click);
		butRemoveProxy.Clicked += new EventHandler (OnRemoveProxy_Click);

		butSaveProxySettings.Clicked += new EventHandler (OnSaveProxySettings_Click);
		butCancelProxySettings.Clicked += new EventHandler (OnCancelProxySettings_Click);
		chkUseProxy.Clicked += new EventHandler (OnUseProxy_Toggle);

		ListStore store = new ListStore(typeof(string), typeof(string));

		lstProxyList.Model = store;
		lstProxyList.HeadersVisible = true;
		lstProxyList.AppendColumn ("Address", new CellRendererText(), "text", 0);
		lstProxyList.AppendColumn ("Port", new CellRendererText(), "text", 1);
	}
	// }}}

	// {{{ ConnectFileSelectionObjects
	private void ConnectFileSelectionObjects(ref Glade.XML gxml, FileAction fa)
	{
		FileSelectForm = (Gtk.FileSelection) gxml.GetWidget("FileSelectForm");
		butFileSelectCancel = (Gtk.Button) gxml.GetWidget("butFileSelectCancel");
		butFileSelectOk = (Gtk.Button) gxml.GetWidget("butFileSelectOk");

		butFileSelectCancel.Clicked += new EventHandler (OnFileSelectCancel_Click);

		switch (fa)
		{
			case FileAction.Open:
				butFileSelectOk.Clicked += new EventHandler (OnFileSelectOk_Click_Open);
				break;
			case FileAction.SaveAs:
				butFileSelectOk.Clicked += new EventHandler (OnFileSelectOk_Click_SaveAs);
				break;
			case FileAction.Pulldata:
				butFileSelectOk.Clicked += new EventHandler (OnFileSelectOk_Click_Pulldata);
				break;
		}	
	}
	// }}}

	// {{{ SetParametersInDataStore
	private void SetParametersInDataStore()
	{
		ListStore store = (ListStore) lstParams.Model;	
		TreeIter iter;
		GlobalDS.FormParam fp = new GlobalDS.FormParam();

		if(store.GetIterFirst(out iter))
		{

			fp.Name = (string) store.GetValue(iter, 0);
			fp.DefaultValue = (string) store.GetValue(iter, 1);
			fp.Injectable = (bool) store.GetValue(iter, 2);

			_SaveInfo.AddFormParameter(fp);

			while (store.IterNext(ref iter))
			{
				fp = new GlobalDS.FormParam();

				fp.Name = (string) store.GetValue(iter, 0);
				fp.DefaultValue = (string) store.GetValue(iter, 1);
				fp.Injectable = (bool) store.GetValue(iter, 2);

				_SaveInfo.AddFormParameter(fp);
			}
		}

	}
	// }}}

	// {{{ SetCookiesInDataStore
	private void SetCookiesInDataStore()
	{
		ListStore store = (ListStore) lstCookies.Model;	
		TreeIter iter;
		StringDictionary CookieList = new StringDictionary();

		string CookieName, CookieValue;

		if(store.GetIterFirst(out iter))
		{

			CookieName = (string) store.GetValue(iter, 0);
			CookieValue = (string) store.GetValue(iter, 1);
			CookieList[CookieName] = CookieValue;

			while (store.IterNext(ref iter))
			{
				CookieName = (string) store.GetValue(iter, 0);
				CookieValue = (string) store.GetValue(iter, 1);
				CookieList[CookieName] = CookieValue;
			}

			_SaveInfo.Cookies = CookieList;
		}

	}
	// }}}

	// {{{ LoadParamsFromStore
	private void LoadParamsFromStore()
	{
		ListStore store = (ListStore) lstParams.Model;

		store.Clear();

		foreach (DictionaryEntry den in _SaveInfo.ParameterTable)
		{
			GlobalDS.FormParam fp = (GlobalDS.FormParam) den.Value;
			store.AppendValues(fp.Name, fp.DefaultValue, fp.Injectable);
		}
		lstParams.Model = store;
	}
	// }}}

	// {{{ LoadCookiesFromStore
	private void LoadCookiesFromStore()
	{
		ListStore store = (ListStore) lstCookies.Model;

		store.Clear();

		foreach (string CookieName in _SaveInfo.Cookies.Keys)
		{
			store.AppendValues(CookieName, _SaveInfo.Cookies[CookieName]);
		}
		lstParams.Model = store;
	}
	// }}}

	// {{{ PullDataFromTables
	private void PullDataFromTables()
	{
		Hashtable TableSet = new Hashtable();
		ListStore store = (ListStore) lstSelectedFields.Model;
		TreeIter iter;
		string TableName, FieldName;
		long FieldID;
		ArrayList al;

		store.GetIterFirst(out iter);

		if (store.IterIsValid(iter))
		{
			do
			{
				TableName = store.GetValue(iter, 0).ToString(); 
				FieldName = store.GetValue(iter, 1).ToString(); 
				FieldID = (int) store.GetValue(iter, 2) + 1L;

				if (!TableSet.ContainsKey(TableName))
				{
					al = new ArrayList();
					TableSet.Add(TableName, al);
				}

				al = (ArrayList) TableSet[TableName];
				al.Add(FieldID);
				TableSet[TableName] = al;

			}while (store.IterNext(ref iter));
		}

		int TabCount = 0;
		ArrayList ColAl, TblAl;
		ColAl = new ArrayList();
		TblAl = new ArrayList();

		foreach (string tab in TableSet.Keys)
		{
			long[] ColumnList = (long[]) ((ArrayList) TableSet[tab]).ToArray(typeof(long));
			TblAl.Add(_SaveInfo.GetTableFromName(tab));
			ColAl.Add( (long[]) ((ArrayList) TableSet[tab]).ToArray(typeof(long)));

			TabCount++;	
		}
		_SaveInfo.TargetAttackVector.PullDataFromTable((GlobalDS.Table[]) TblAl.ToArray(typeof(GlobalDS.Table)), (long[][]) ColAl.ToArray(typeof(long[])), txtPullDataXml.Text);
	}
	// }}}

	// Events

	// {{{ OnQuit_Click
	public void OnQuit_Click(object sender, EventArgs args)
	{
		Application.Quit();
	}
	// }}}

	// {{{ OnGetFieldInfo_Click
	public void OnLoadFieldInfo_Click(object sender, EventArgs args)
	{
		try
		{
			if (_SaveInfo.TargetAttackVector != null && _SaveInfo.TableList != null)
			{
				// This will depend on all the combolist
				TreePath TablePath;
				TreeViewColumn TableColumn;
				lstSchema.GetCursor(out TablePath, out TableColumn);

				// Yeah, I know hardcoding is bad, but this makes the most sense atm
				if (TablePath.IsDescendant(new TreePath("0:1")))
				{
					while (TablePath.Depth > 3)
					{
						TablePath.Up();
					}

					TreeIter TableIter;
					TreeStore store = (TreeStore) lstSchema.Model;

					store.GetIter(out TableIter, TablePath);

					string TableName = store.GetValue(TableIter, 0).ToString();

					LoadFieldDataFromTableName(TableName, TableIter, false);
				}

			}
		}
		catch (WebException wex)
		{
			MessageDialog md = new MessageDialog (Absinthe,0, MessageType.Error, ButtonsType.Close, wex.Status.ToString());
			int result = md.Run();
		}
	}
	// }}}

	// {{{ OnGetUsername_Click
	public void OnGetUsername_Click(object sender, EventArgs args)
	{
		ThreadedSub a = new ThreadedSub(RetrieveUsernameFromDatabase);
		a.BeginInvoke(null, new object());
	}
	// }}}

	// {{{ OnGetTableInfo_Click
	public void OnLoadTableInfo_Click(object sender, EventArgs args)
	{
		ThreadedSub a = new ThreadedSub(RetrieveTableInfoFromDatabase);
		a.BeginInvoke(null, new object());
	}
	// }}}

	// {{{ OnInjectable_Click
	public void OnInjectable_Click(object sender, EventArgs args)
	{
		chkTreatAsString.Sensitive = chkInjectable.Active;
		chkTreatAsString.Active = false;
	}
	// }}}

	// {{{ OnScan_Click
	public void OnScan_Click(object sender, EventArgs args)
	{
		//InitializeAttackVectors();	
		ThreadedSub a = new ThreadedSub(InitializeAttackVectors);
		a.BeginInvoke(null, new object());
	}
	// }}}

	// {{{ OnConfigProxy_Click
	public void OnConfigProxy_Click(object sender, EventArgs args)
	{
		Glade.XML gxml = new Glade.XML (null, "gtkabsinthe.glade", "ProxyConfig", null);
		//Glade.XML gxml = new Glade.XML ("gtkabsinthe.glade", "ProxyConfig", null);
		Glade.XML gxml2 = new Glade.XML (null, "gtkabsinthe.glade", "AboutForm", null);
		//Glade.XML gxml2 = new Glade.XML ("gtkabsinthe.glade", "AboutForm", null);

		ConnectProxyConfigObjects(ref gxml);

		ProxyConfig.Visible = true;

		chkUseProxy.Active = _AppSettings.ProxyInUse;

		LoadProxyListView();

		txtProxyAddress.Sensitive = _AppSettings.ProxyInUse;
		txtProxyPort.Sensitive = _AppSettings.ProxyInUse;
		lstProxyList.Sensitive = _AppSettings.ProxyInUse;
		butAddProxy.Sensitive = _AppSettings.ProxyInUse;
		butRemoveProxy.Sensitive = _AppSettings.ProxyInUse;
	}
	// }}}

	// {{{ LoadProxyListView
	private void LoadProxyListView()
	{
		Queue ProxyList = _AppSettings.ProxyQueue;
		ListStore store = (ListStore) lstProxyList.Model;

		if (ProxyList != null)
			foreach (WebProxy wp in ProxyList)
			{
				store.AppendValues(wp.Address.Host, wp.Address.Port);
			}
	}
	// }}}

	// {{{ OnAddParam_Click
	public void OnAddParam_Click(object sender, EventArgs args)
	{
		AddParamToListView();
	}
	// }}}

	// {{{ OnAddHeader_Click
	public void OnAddHeader_Click(object sender, EventArgs args)
	{
		AddCookieToListView();
	}
	// }}}

	// {{{ AddCookieToListView
	private void AddCookieToListView()
	{
		ListStore store = (ListStore) lstCookies.Model;
		string CookieName = txtParamName.Text;
		string CookieValue = txtParamValue.Text;

		store.AppendValues(CookieName, CookieValue);
		lstCookies.Model = store;

		txtParamName.Text = String.Empty;
		txtParamValue.Text = String.Empty;
		chkInjectable.Active = false;
		chkTreatAsString.Active = false;
	}
	// }}}

	// {{{ AddParamToListView
	private void AddParamToListView()
	{
		ListStore store = (ListStore) lstParams.Model;
		string ParamName = txtParamName.Text;
		string ParamValue = txtParamValue.Text;

		if (ParamName.Trim().Length == 0 && ParamValue.Trim().Length == 0) return;

		store.AppendValues(ParamName,ParamValue, chkInjectable.Active, chkTreatAsString.Active);
		lstParams.Model = store;

		txtParamName.Text = String.Empty;
		txtParamValue.Text = String.Empty;
		chkInjectable.Active = false;
		chkTreatAsString.Active = false;
	}
	// }}}

	// {{{ OnEditParam_Click
	public void OnEditParam_Click(object sender, EventArgs args)
	{
		TreeIter iter = new TreeIter ();
		TreeModel model = lstParams.Model;
		ListStore store = (ListStore) lstParams.Model;	
		TreeView lstActive;
		lstActive = lstParams;
		if (tabParams.CurrentPage == 1) 
		{
			store = (ListStore) lstCookies.Model;
			model = lstCookies.Model;
			lstActive = lstCookies;
		}


		if (lstActive.Selection.GetSelected (out model, out iter)) {
			txtParamName.Text = (string) model.GetValue(iter, 0);
			txtParamValue.Text = (string) model.GetValue(iter, 1);
			if (tabParams.CurrentPage == 0)
			{
				chkInjectable.Active = (bool) model.GetValue(iter, 2);
			}
			else
			{
				chkInjectable.Active = false;
			}
			store.Remove(ref iter);
		}
	}
	// }}}

	// {{{ OnDeleteParam_Click
	public void OnDeleteParam_Click(object sender, EventArgs args)
	{
		TreeIter iter = new TreeIter ();
		TreeModel model = lstParams.Model;
		ListStore store = (ListStore) lstParams.Model;	
		TreeView lstActive;

		lstActive = lstParams;
		if (tabParams.CurrentPage == 1) 
		{
			store = (ListStore) lstCookies.Model;
			model = lstCookies.Model;
			lstActive = lstCookies;
		}

		if (lstActive.Selection.GetSelected (out model, out iter)) {
			store.Remove(ref iter);
		}
	}
	// }}}

	// {{{ OnSaveProxySettings_Click
	public void OnSaveProxySettings_Click(object sender, EventArgs args)
	{	
		ListStore store = (ListStore) lstProxyList.Model;	
		TreeIter iter;

		string ProxyHostName;
		int ProxyPort;

		if(store.GetIterFirst(out iter))
		{

			ProxyHostName = (string) store.GetValue(iter, 0);
			ProxyPort = Int32.Parse((string) store.GetValue(iter, 1));

			_AppSettings.AddProxy(ProxyHostName, ProxyPort);

			while (store.IterNext(ref iter))
			{
				ProxyHostName = (string) store.GetValue(iter, 0);
				ProxyPort = Int32.Parse((string) store.GetValue(iter, 1));

				_AppSettings.AddProxy(ProxyHostName, ProxyPort);
			}
		}

		ProxyConfig.Destroy();
	}
	// }}}

	// {{{ OnAddProxy_Click
	public void OnAddProxy_Click(object sender, EventArgs args)
	{
		//ProxyConfig.Destroy();
		if(Regex.IsMatch(txtProxyPort.Text, @"^\d*$"))
		{
			//	_AppSettings.SetProxyInfo(txtProxyAddress.Text, Int32.Parse(txtProxyPort.Text), chkUseProxy.Active);
			ListStore store = (ListStore) lstProxyList.Model;
			store.AppendValues(txtProxyAddress.Text, txtProxyPort.Text);
			txtProxyAddress.Text = String.Empty;
			txtProxyPort.Text = String.Empty;
		}
		else
		{
			GenerateErrorDialog("The proxy port you have selected is invalid.");
		}
	}
	// }}}

	// {{{ OnRemoveProxy_Click
	public void OnRemoveProxy_Click(object sender, EventArgs args)
	{
		TreeIter iter = new TreeIter ();
		TreeModel model = lstProxyList.Model;
		ListStore store = (ListStore) lstProxyList.Model;	

		txtProxyAddress.Text = String.Empty;
		txtProxyPort.Text = String.Empty;
		lstProxyList.Selection.GetSelected(out model, out iter);
		store.Remove(ref iter);
	}
	// }}}

	// {{{ OnCancelProxySettings_Click
	public void OnCancelProxySettings_Click(object sender, EventArgs args)
	{
		ProxyConfig.Destroy();
	}
	// }}}

	// {{{ OnSave_Click
	public void OnSave_Click(object sender, EventArgs args)
	{
		if (_SaveInfo.OutputFile.Length == 0)
		{
			OnSaveAs_Click(sender, args);
			return;
		}

		_SaveInfo.TargetURL = txtTargetURL.Text;
		_SaveInfo.ConnectionMethod = ConnectMethod();

		SetParametersInDataStore();
		SetCookiesInDataStore();

		_SaveInfo.OutputToFile(cboPluginList.Entry.Text);
	}
	// }}}

	// {{{ OnOpen_Click
	public void OnOpen_Click(object sender, EventArgs args)
	{
		Glade.XML gxml = new Glade.XML (null, "gtkabsinthe.glade", "FileSelectForm", null);
		//Glade.XML gxml = new Glade.XML ("gtkabsinthe.glade", "FileSelectForm", null);

		ConnectFileSelectionObjects(ref gxml, FileAction.Open);
		if (FileSelectForm != null) FileSelectForm.Visible = true;
	}
	// }}}

	// {{{ OnMenuOptions_Click
	public void OnMenuOptions_Click(object sender, EventArgs args)
	{
		gtkInjectionOptions OptionsWindow = new gtkInjectionOptions(ref _SaveInfo);
	}
	// }}}

	// {{{ OnAbout_Click
	public void OnAbout_Click(object sender, EventArgs args)
	{
		Glade.XML gxml = new Glade.XML (null, "gtkabsinthe.glade", "AboutForm", null);
		//Glade.XML gxml = new Glade.XML ("gtkabsinthe.glade", "AboutForm", null);

		if (gxml != null)
		{
		
			AboutForm = (Gtk.Window) gxml.GetWidget("AboutForm");
			butAboutOk = (Gtk.Button) gxml.GetWidget("butAboutOk");

			
			AboutImage = (Gtk.Image) gxml.GetWidget("AboutImage");
			AboutImage.Pixbuf = new Pixbuf (null, "log0.gif");
			
			butAboutOk.Clicked += new EventHandler (OnAboutOk_Click);

			if (AboutForm != null) AboutForm.Visible = true;
		}
		
	}
	// }}}

	// {{{ OnNew_Click
	public void OnNew_Click(object sender, EventArgs args)
	{
		// Reset Everything
		((ListStore)lstParams.Model).Clear();
		((ListStore)lstCookies.Model).Clear();
		txtTargetURL.Text = String.Empty;
		optConnectPost.Active = true;
		InitUI();

		// Reset the data store, so there's no lingering info
		_SaveInfo = new DataStore(new GlobalDS.OutputStatusDelegate(SafeOutput));

	}
	// }}}

	// {{{ OnSaveAs_Click
	public void OnSaveAs_Click(object sender, EventArgs args)
	{
		Glade.XML gxml = new Glade.XML (null, "gtkabsinthe.glade", "FileSelectForm", null);
		//Glade.XML gxml = new Glade.XML ("gtkabsinthe.glade", "FileSelectForm", null);

		ConnectFileSelectionObjects(ref gxml, FileAction.SaveAs);
		if (FileSelectForm != null) FileSelectForm.Visible = true;
	}
	// }}}
	
	// {{{ OnAboutOk_Click
	public void OnAboutOk_Click(object sender, EventArgs args)
	{
		AboutForm.Destroy();
	}
	// }}}

	// {{{ OnFileSelectCancel_Click
	public void OnFileSelectCancel_Click(object sender, EventArgs args)
	{
		FileSelectForm.Destroy();
	}
	// }}}

	// {{{ OnAddDlField_Click
	public void OnAddDlField_Click(object sender, EventArgs args)
	{
		try
		{
			if (_SaveInfo.TargetAttackVector != null && _SaveInfo.TableList != null)
			{
				// This will depend on all the combolist
				TreePath TablePath;
				TreeViewColumn TableColumn;
				TreeIter iter;

				lstAvailFields.GetCursor(out TablePath, out TableColumn);

				if (TablePath == null || lstAvailFields == null) return;

				lstAvailFields.Model.GetIter(out iter, TablePath);

				string TableName, FieldName;
				int FieldID;
				
				if (lstAvailFields.Model == null) return;
				ListStore astore = (ListStore) lstAvailFields.Model;
				TableName = astore.GetValue(iter, 0).ToString(); 
				FieldName = astore.GetValue(iter, 1).ToString(); 
				FieldID = (int) astore.GetValue(iter, 2); 

				astore.Remove(ref iter);

				ListStore fstore = (ListStore)lstSelectedFields.Model;
				fstore.AppendValues(TableName,FieldName,FieldID);

			}
		}
		catch (System.NullReferenceException nre)
		{
			// Ignore
		}
	}
	// }}}

	// {{{ OnRemoveDlField_Click
	public void OnRemoveDlField_Click(object sender, EventArgs args)
	{
		try
		{
			if (_SaveInfo.TargetAttackVector != null && _SaveInfo.TableList != null)
			{
				// This will depend on all the combolist
				TreePath TablePath;
				TreeViewColumn TableColumn;
				TreeIter iter;

				lstSelectedFields.GetCursor(out TablePath, out TableColumn);
				lstSelectedFields.Model.GetIter(out iter, TablePath);

				string TableName, FieldName;
				int FieldID;
				
				ListStore astore = (ListStore) lstSelectedFields.Model;
				TableName = astore.GetValue(iter, 0).ToString(); 
				FieldName = astore.GetValue(iter, 1).ToString(); 
				FieldID = (int) astore.GetValue(iter, 2);
				astore.Remove(ref iter);

				ListStore fstore = (ListStore)lstAvailFields.Model;
				fstore.AppendValues(TableName,FieldName,FieldID);

			}
		}
		catch (System.NullReferenceException nre)
		{
			// Ignore
		}
	}
	// }}}

	// {{{ OnBrowseDatapull_Click
	public void OnBrowseDatapull_Click(object sender, EventArgs args)
	{
		Glade.XML gxml = new Glade.XML (null, "gtkabsinthe.glade", "FileSelectForm", null);
		//Glade.XML gxml = new Glade.XML ("gtkabsinthe.glade", "FileSelectForm", null);

		ConnectFileSelectionObjects(ref gxml, FileAction.Pulldata);
		if (FileSelectForm != null) FileSelectForm.Visible = true;
	}
	// }}}

	// {{{ OnFileSelectOk_Click_Open
	public void OnFileSelectOk_Click_Open(object sender, EventArgs args)
	{
		GlobalDS.OutputStatusDelegate safeout = new GlobalDS.OutputStatusDelegate(SafeOutput);

		try
		{
			_SaveInfo.LoadXmlFile(FileSelectForm.Filename, safeout, _AppSettings.ProxyQueue);
		}
		catch (DataStore.InvalidDataFileException idfe)
		{
			//MsgBox(idfe.Message());
			MessageDialog md = new MessageDialog (FileSelectForm,DialogFlags.DestroyWithParent, MessageType.Error, ButtonsType.Close, idfe.Message());
			int result = md.Run();
		}

		if (_SaveInfo.ConnectionMethod.ToUpper().Equals("POST")) { optConnectPost.Active = true; }
		else { optConnectGet.Active = true; }

		txtTargetURL.Text = _SaveInfo.TargetURL;

		LoadParamsFromStore();

		optBlindInjection.Active = false; optErrorBasedInjection.Active = false;
		chkTerminateQuery.Active = _SaveInfo.TerminateQuery;

		if (_SaveInfo.LoadedPluginName != null)
		{
			LoadPluginSelection(_SaveInfo.LoadedPluginName);
		}
			
		if (_SaveInfo.TargetAttackVector != null)
		{
			switch(_SaveInfo.TargetAttackVector.ExploitType)
			{
				case GlobalDS.ExploitType.ErrorBasedTSQL:
					optErrorBasedInjection.Active = true;
					break;
				case GlobalDS.ExploitType.BlindTSQLInjection:
					optBlindInjection.Active = true;
					break;
			}
		}

		LoadSchemaData();

		FileSelectForm.Destroy();
	}
	// }}}

	private void LoadPluginSelection(string PluginText)
	{
		cboPluginList.Entry.Text = PluginText;
	}

	// {{{ LoadSchemaData
	private void LoadSchemaData()
	{
		if (_SaveInfo.Username.Length > 0)
		{
			AdjustDbLoginText();
		}

		if (_SaveInfo.TableList != null)
		{
			SafeOutput("one");
			LoadTableInfoList();
			SafeOutput("two");

			TreePath TablePath = new TreePath("0:1");
			TreeIter TableIter, NameIter;
			TreeStore store = (TreeStore) lstSchema.Model;
			SafeOutput("three");

			store.GetIter(out TableIter, TablePath);
			SafeOutput("TableIter: #" + TableIter.Stamp.ToString());

			SafeOutput("N Children: #" + store.IterNChildren().ToString());
			for (int i=0; i < store.IterNChildren(TableIter); i++)
			{
				TreePath TableNamePath = new TreePath("0:1:" + i.ToString());
				store.GetIter(out NameIter, TableNamePath);
				string TableName = store.GetValue(NameIter, 0).ToString();
				SafeOutput("Table Name: " + TableName);

				SafeOutput("NameIter: #" + NameIter.Stamp.ToString());
				//if (NameIter != null)
				//{
					LoadFieldDataFromTableName(TableName, NameIter, true);
				//}
				//else
				//{
				//	SafeOutput("Null NameIter!!");
				//}
			}

			butLoadFieldInfo.Visible = true; // Explicit in case it wasn't called properly
		}
	}

	// }}}

	// {{{ OnFileSelectOk_Click_SaveAs
	public void OnFileSelectOk_Click_SaveAs(object sender, EventArgs args)
	{
		_SaveInfo.OutputFile = FileSelectForm.Filename;
		OnSave_Click(sender, args);
		FileSelectForm.Destroy();
	}
	// }}}

	// {{{ OnFileSelectOk_Click_Pulldata
	public void OnFileSelectOk_Click_Pulldata(object sender, EventArgs args)
	{
		txtPullDataXml.Text = FileSelectForm.Filename;
		FileSelectForm.Destroy();
	}
	// }}}

	// {{{ OnPullData_Click
	public void OnPullData_Click(object sender, EventArgs args)
	{
		//Application.Quit();
		if (txtPullDataXml.Text.Length > 0)
		{
			PullDataFromTables();		
		}
		else
		{
			myStatusBar.Push(myStatusBar.GetContextId("regular"), "Please select a file!");
		}
	}
	// }}}

	// {{{ OnAppendExtraQuery_Click
	private void OnAppendExtraQuery_Click(object sender, EventArgs args)
	{
		chkTerminateQuery.Active = false;

		if (chkAppendExtraQuery.Active)
		{
			txtAppendExtraQuery.Visible = true;
			lblAppendExtraQuery.Visible = true;
		}
		else
		{
			txtAppendExtraQuery.Text = String.Empty;
			txtAppendExtraQuery.Visible = false;
			lblAppendExtraQuery.Visible = false;
		}
	}
	// }}}

	// {{{ OnTerminateQuery_Click
	private void OnTerminateQuery_Click(object sender, EventArgs args)
	{
		chkAppendExtraQuery.Active = false;

		txtAppendExtraQuery.Visible = false;
		txtAppendExtraQuery.Text = String.Empty;
		lblAppendExtraQuery.Visible = false;
	}
	// }}}

	// 
	

}
