/*****************************************************************************
   gtkAbsinthe - Front End for the Absinthe Core Library
   This software is Copyright (C) 2004  nummish, 0x90.org

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

using System;
using System.Text.RegularExpressions;

using Gtk;
using Glade;
using GLib;
using Gdk;
using GtkSharp;

public class gtkInjectionOptions
{
	Glade.XML LocalGladeXml;
	[Glade.Widget] Gtk.Window WindowObject;
	[Glade.Widget] Gtk.Entry txtTolerance;
	[Glade.Widget] Gtk.Entry txtFilterDelimiter;
	[Glade.Widget] Gtk.Entry txtThrottle;
	[Glade.Widget] Gtk.HScale sldThrottleSlider;

	[Glade.Widget] Gtk.Dialog dlgError;
	[Glade.Widget] Gtk.Label lblErrorMsg;
	[Glade.Widget] Gtk.Button okbutton1;

	Absinthe.Core.DataStore _Storage;

	public gtkInjectionOptions(ref Absinthe.Core.DataStore Storage)
	{
		LocalGladeXml = new Glade.XML (null, "gtkabsinthe.glade", "InjectionOptions", null);
		LocalGladeXml.Autoconnect(this);
		WindowObject = (Gtk.Window) LocalGladeXml.GetWidget("InjectionOptions"); 

		_Storage = Storage;

		InitUI();
	}

	private void InitUI()
	{
		txtTolerance.Text = (_Storage.FilterTolerance * 100).ToString();
		if (_Storage.FilterDelimiter.Equals(System.Environment.NewLine))
		{
			txtFilterDelimiter.Text = String.Empty;
		}
		else
		{
			txtFilterDelimiter.Text = _Storage.FilterDelimiter;
		}
		txtThrottle.Text = _Storage.ThrottleValue.ToString();
		sldThrottleSlider.Value = _Storage.ThrottleValue;
	}

	public void butCancel_Click(object sender, EventArgs args)
	{
		WindowObject.Destroy();
	}

	public void butOK_Click(object sender, EventArgs args)
	{
		// this should allow decimals too
		//if (Regex.IsMatch(txtTolerance.Text, @"^\d*[\.\d*]?$")) 
		if (Regex.IsMatch(txtTolerance.Text, @"^[+]?\d+(\.\d+)?$") == false) 
		{
			GenerateErrorDialog("You have entered an invalid tolerance.");
			return;
		}
				
		_Storage.FilterTolerance = Single.Parse(txtTolerance.Text) / 100;
		_Storage.FilterDelimiter = txtFilterDelimiter.Text;

		int ThrottleValue;
		ThrottleValue = Convert.ToInt32(sldThrottleSlider.Value);

		if (ThrottleValue == 0) 
		{
			ThrottleValue = Int32.Parse(txtThrottle.Text);
		}
		else
		{	
			ThrottleValue *= -100;
		}
		
		_Storage.ThrottleValue = ThrottleValue;

		WindowObject.Destroy();
	}

	// {{{ GenerateErrorDialog
	private void GenerateErrorDialog(string ErrorMessage)
	{
		
		Glade.XML gxml = new Glade.XML (null, "gtkabsinthe.glade", "dlgError", null);
		dlgError = (Gtk.Dialog) gxml.GetWidget("dlgError");
		lblErrorMsg = (Gtk.Label) gxml.GetWidget("lblErrorMsg");
		lblErrorMsg.Text = ErrorMessage;
		okbutton1 = (Gtk.Button) gxml.GetWidget("okbutton1");
		okbutton1.Clicked += new EventHandler (ErrorOk_Click);
	}
	// }}}

	// {{{ ErrorOk_Click
	public void ErrorOk_Click(object sender, EventArgs args)
	{
		dlgError.Destroy();
	}
	// }}}

	public void OntxtThrottle_Change(object sender, EventArgs args)
	{
		sldThrottleSlider.Value = 1;
	}

	public void OnThrottleSlider_Change(object sender, EventArgs args)
	{
		// The Textchange likes to reset the slider
		double ThrottleVal = sldThrottleSlider.Value;
		txtThrottle.Text = "0";
		sldThrottleSlider.Value = ThrottleVal;
	}
}
